searchData={"content_type":"text/markdown","items":[{"doc":"Venomous is a wrapper around erlport python Ports, designed to simplify concurrent use. It focuses dynamic extensibility, like spawning, reusing and killing processes on demand. Furthermore, unused processes get automatically killed by scheduled process which can be configured inside config.exs. Venomous core functions capture and handle :EXIT calls ensuring that all python process die with it and do not continue their execution.\n\nThe core concept revolves around \"Snakes\" which represent Python worker processes. These `Venomous.SnakeWorker` are managed and supervised with `Venomous.SnakeManager` GenServer to allow concurrent and efficient execution of Python code. The `Snakes` pids and python pids are stored inside `:ets` table and the Processes are handled by `DymanicSupervisor` called `Venomous.SnakeSupervisor`. The unused `Snakes` get automatically killed by `SnakeManager` depending on the given configuration.","ref":"Venomous.html","title":"Venomous","type":"module"},{"doc":"- `python/2`: The primary function to execute a Python function. It retrieves a Snake (Python worker process) and runs the specified Python function using the arguments provided in a `SnakeArgs` struct. If no ready Snakes are available, a new one is spawned. If max_children is reached it will return an error with appropriate message.\n- `python!/2` | `python!/1`: Will wait until any `Venomous.SnakeWorker` is freed, requesting it with the given interval.","ref":"Venomous.html#module-main-functionality","title":"Main Functionality - Venomous","type":"module"},{"doc":"Venomous consists of several key components:\n\n- `Venomous.SnakeWorker`: Manages the execution of Python processes.\n- `Venomous.SnakeSupervisor`: A DynamicSupervisor that oversees the SnakeWorkers.\n- `Venomous.SnakeManager`: A GenServer that coordinates the SnakeWorkers and handles operations like spawning, retrieval and cleanup.","ref":"Venomous.html#module-architecture","title":"Architecture - Venomous","type":"module"},{"doc":"The behavior and management of Snakes can be configured through the following options inside :venomous :snake_manager config key:\n\n- `erlport_encoder: %{module: atom(), func: atom(), args: list(any())}`: Optional :erlport encoder/decoder python function for converting types. This function is applied to every unnamed python process started by SnakeManager. For more information see [Handling Erlport API](PYTHON.md)\n- `snake_ttl_minutes: non_neg_integer()`: Time-to-live for a Snake in minutes. Default is 15 min.\n- `perpetual_workers: non_neg_integer()`: Number of Snakes to keep alive perpetually. Default is 10.\n- `cleaner_interval: non_neg_integer()`: Interval in milliseconds for cleaning up inactive Snakes. Default is 60_000 ms.","ref":"Venomous.html#module-configuration-options","title":"Configuration Options - Venomous","type":"module"},{"doc":"- `list_alive_snakes/0`: Returns a list of :ets table containing currently alive Snakes.\n- `clean_inactive_snakes/0`: Manually clears inactive Snakes depending on their ttl and returns the number of Snakes cleared.\n- `slay_python_worker/2`: Kills a specified Python worker process and its SnakeWorker. :brutal can be specified as option, which will `kill -9` the os process of python which prevents the code from executing until it finalizes or goes through iteration.","ref":"Venomous.html#module-auxiliary-functions","title":"Auxiliary Functions - Venomous","type":"module"},{"doc":"Clears inactive snakes manually, returns number of snakes cleared.","ref":"Venomous.html#clean_inactive_snakes/0","title":"Venomous.clean_inactive_snakes/0","type":"function"},{"doc":"Retrieves x amount of ready snakes. In case of hitting max_children cap, stops and returns all available snakes.","ref":"Venomous.html#get_snakes_ready/1","title":"Venomous.get_snakes_ready/1","type":"function"},{"doc":"- amount of snakes to retrieve","ref":"Venomous.html#get_snakes_ready/1-parameters","title":"Parameters - Venomous.get_snakes_ready/1","type":"function"},{"doc":"- A list of tuples `{pid, pid}`","ref":"Venomous.html#get_snakes_ready/1-returns","title":"Returns - Venomous.get_snakes_ready/1","type":"function"},{"doc":"Returns list of :ets table containing alive snakes","ref":"Venomous.html#list_alive_snakes/0","title":"Venomous.list_alive_snakes/0","type":"function"},{"doc":"Wrapper for python workers\nTries to retrieve `Venomous.SnakeWorker` which then runs given function inside module with args. In case of failure will return {:error, message}.\nIn case :EXIT happens, it will kill python worker/process and exit(reason)","ref":"Venomous.html#python/2","title":"Venomous.python/2","type":"function"},{"doc":"- %SnakeArgs{} struct of :module, :func, :args \n  - opts \\ []","ref":"Venomous.html#python/2-parameters","title":"Parameters - Venomous.python/2","type":"function"},{"doc":"- `:python_timeout` ms timeout. Kills python OS process on timeout. Default: 15_000\n- `:kill_python_on_exception` Should python process be killed on exception. Should be fine to set to `true` if your are sure the Python process won't exit(). In case exit() happens and it's set to true, Python will never reply, waiting till timeout. Default: false","ref":"Venomous.html#python/2-opts","title":"Opts - Venomous.python/2","type":"function"},{"doc":"- any() | {error: :timeout} | {error: any()} retrieves output of python function or error","ref":"Venomous.html#python/2-returns","title":"Returns - Venomous.python/2","type":"function"},{"doc":"If no Snake is available will continue requesting it with the given interval until any gets freed or receives :EXIT signal","ref":"Venomous.html#python!/2","title":"Venomous.python!/2","type":"function"},{"doc":"- `:retrieve_interval` ms to wait before requesting snake again Default: 200\n- `:python_timeout` ms timeout. Kills python OS process on timeout. Default: 15_000\n- `:kill_python_on_exception` Should python process be killed on exception. Should be fine to set to `true` if your are sure the Python process won't exit(). In case exit() happens and it's set to true, Python will never reply, waiting till timeout. Default: false","ref":"Venomous.html#python!/2-opts","title":"Opts - Venomous.python!/2","type":"function"},{"doc":"Retrieves ready SnakeWorker and python pids.\nIf all processes are busy and exceeds max_children will return {:retrieve_error, message}.","ref":"Venomous.html#retrieve_snake/0","title":"Venomous.retrieve_snake/0","type":"function"},{"doc":"- A tuple `{pid, pid}` containing the process IDs of the SnakeWorker and python processes. In case of error `{:retrieve_error, message}`","ref":"Venomous.html#retrieve_snake/0-returns","title":"Returns - Venomous.retrieve_snake/0","type":"function"},{"doc":"Retrieves ready SnakeWorker and python pids.\nThe worker is then set to :busy until its ran with snake_run(), preventing it from getting removed automatically or used by other process\nIf all processes are busy and exceeds max_children will wait for interval ms and try again. Traps the exit signals, to safely escape loop.","ref":"Venomous.html#retrieve_snake!/1","title":"Venomous.retrieve_snake!/1","type":"function"},{"doc":"- interval: The time to wait in milliseconds before retrying. Default is `@wait_for_snake_interval`.","ref":"Venomous.html#retrieve_snake!/1-parameters","title":"Parameters - Venomous.retrieve_snake!/1","type":"function"},{"doc":"- A tuple `{pid, pid}` containing the process IDs of the SnakeWorker and python processes.","ref":"Venomous.html#retrieve_snake!/1-returns","title":"Returns - Venomous.retrieve_snake!/1","type":"function"},{"doc":"Kills python process and its SnakeWorker\n:brutal also kills the OS process of python, ensuring the process does not continue execution.","ref":"Venomous.html#slay_python_worker/2","title":"Venomous.slay_python_worker/2","type":"function"},{"doc":"- SnakeWorker pid, Python pid tuple\n  - a Way to kill process. default: :brutal additionally kills with kill -9 ensuring the python does not execute further","ref":"Venomous.html#slay_python_worker/2-parameters","title":"Parameters - Venomous.slay_python_worker/2","type":"function"},{"doc":":ok","ref":"Venomous.html#slay_python_worker/2-returns","title":"Returns - Venomous.slay_python_worker/2","type":"function"},{"doc":"Runs given %SnakeArgs{} inside given Snake pids. Takes in a python_timeout which brutally kills the python process after surpassing it.\nTraps exit and awaits signals [:SNAKE_DONE, :SNAKE_ERROR, :EXIT]\nIn case of an exit, brutally kills the python process ensuring it doesn't get executed any further.","ref":"Venomous.html#snake_run/3","title":"Venomous.snake_run/3","type":"function"},{"doc":"- %SnakeArgs{} struct of :module, :func, :args \n  - {pid, pypid} tuple containing SnakeWorker pid and python pid\n  - opts \\ []","ref":"Venomous.html#snake_run/3-parameters","title":"Parameters - Venomous.snake_run/3","type":"function"},{"doc":"- `:python_timeout` ms timeout. Kills python OS process on timeout. Default: 15_000\n- `:kill_python_on_exception` Should python process be killed on exception. Should be fine to set to `true` if your are sure the Python process won't exit(). In case exit() happens and it's set to true, Python will never reply, waiting till timeout. Default: false","ref":"Venomous.html#snake_run/3-opts","title":"Opts - Venomous.snake_run/3","type":"function"},{"doc":"- any() | {error: :timeout} | %SnakeError{} retrieves output of python function or error","ref":"Venomous.html#snake_run/3-returns","title":"Returns - Venomous.snake_run/3","type":"function"},{"doc":"This module initializes the application supervision tree.\nIt starts the supervisor for managing SnakeManager process with the given Application config.","ref":"Venomous.Application.html","title":"Venomous.Application","type":"module"},{"doc":"","ref":"Venomous.Application.html#start/2","title":"Venomous.Application.start/2","type":"function"},{"doc":"A module to handle arguments for a Python function call.\n\nThis module defines a struct for storing the module name, function name,\nand the list of arguments that can be used to call a Python function.","ref":"Venomous.SnakeArgs.html","title":"Venomous.SnakeArgs","type":"module"},{"doc":"Builds SnakeArgs struct from params","ref":"Venomous.SnakeArgs.html#from_params/3","title":"Venomous.SnakeArgs.from_params/3","type":"function"},{"doc":"- module atom() of python module ex. :builtins\n  - function atom() of function from given module ex. :sum\n  - args list(any()) list of arguments for function ex. [ [1,2,3,4,5] ]","ref":"Venomous.SnakeArgs.html#from_params/3-parameters","title":"Parameters - Venomous.SnakeArgs.from_params/3","type":"function"},{"doc":"%SnakeArgs{}","ref":"Venomous.SnakeArgs.html#from_params/3-returns","title":"Returns - Venomous.SnakeArgs.from_params/3","type":"function"},{"doc":"","ref":"Venomous.SnakeArgs.html#t:t/0","title":"Venomous.SnakeArgs.t/0","type":"type"},{"doc":"A module to handle errors raised during Python function calls.\n\nThis module defines a struct for storing exception information,\nincluding the exception itself, the error message, and the backtrace.","ref":"Venomous.SnakeError.html","title":"Venomous.SnakeError","type":"module"},{"doc":"","ref":"Venomous.SnakeError.html#t:t/0","title":"Venomous.SnakeError.t/0","type":"type"},{"doc":"Manager for brave 🐍 workers\n\nThis module manages the snake workers, ensuring that inactive workers are cleaned up periodically.\n  \nMain call `:get_ready_snake` retrieves/spawns a `Venomous.SnakeWorker` with :retrieved status.\nWorkers with status :ready, :retrieved, are considered inactive and will be cleared up by main process loop running `:clean_inactive_workers` if they exceed their given TTL","ref":"Venomous.SnakeManager.html","title":"Venomous.SnakeManager","type":"module"},{"doc":"The following configurations are retrieved from :venomous :snake_manager Application env:\n\n- `erlport_encoder: %{module: atom(), func: atom(), args: list(any())}`: Optional :erlport encoder/decoder python function for converting types. The function can also provide any callbacks from :erlport documentation like .cast() message handler\n- `snake_ttl_minutes: non_neg_integer()`: Time-to-live for a Snake in minutes. Default is 15 min.\n- `perpetual_workers: non_neg_integer()`: Number of Snakes to keep alive perpetually. Default is 10.\n- `cleaner_interval: non_neg_integer()`: Interval in milliseconds for cleaning up inactive Snakes. Default is 60_000 ms.\n\nDefaults are provided in case these configurations are not set:\n\n- Default encoder: none.\n- Default time-to-live for a worker: 15 minutes.\n- Default number of perpetual workers: 10.\n- Default interval for cleaning inactive workers: 60,000 milliseconds.","ref":"Venomous.SnakeManager.html#module-configuration","title":"Configuration - Venomous.SnakeManager","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Venomous.SnakeManager.html#child_spec/1","title":"Venomous.SnakeManager.child_spec/1","type":"function"},{"doc":"","ref":"Venomous.SnakeManager.html#get_snake_worker_status/2","title":"Venomous.SnakeManager.get_snake_worker_status/2","type":"function"},{"doc":"","ref":"Venomous.SnakeManager.html#handle_continue/2","title":"Venomous.SnakeManager.handle_continue/2","type":"function"},{"doc":"","ref":"Venomous.SnakeManager.html#init/1","title":"Venomous.SnakeManager.init/1","type":"function"},{"doc":"","ref":"Venomous.SnakeManager.html#start_link/1","title":"Venomous.SnakeManager.start_link/1","type":"function"},{"doc":"DynamicSupervisor for Snakes\nThe maximum cap of `Venomous.SnakeWorker` is defined with :max_children option\n> #### Warning {: .warning}\n>\n> The maximum amount of python processes is capped at your systems maximum number of open file-descriptors\n> `ulimit -n` to check your limit","ref":"Venomous.SnakeSupervisor.html","title":"Venomous.SnakeSupervisor","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Venomous.SnakeSupervisor.html#child_spec/1","title":"Venomous.SnakeSupervisor.child_spec/1","type":"function"},{"doc":"Deploys snake with given opts Map containing :erlport encoder/decoder module, func and args","ref":"Venomous.SnakeSupervisor.html#deploy_snake_worker/1","title":"Venomous.SnakeSupervisor.deploy_snake_worker/1","type":"function"},{"doc":"","ref":"Venomous.SnakeSupervisor.html#init/1","title":"Venomous.SnakeSupervisor.init/1","type":"function"},{"doc":"","ref":"Venomous.SnakeSupervisor.html#start_link/1","title":"Venomous.SnakeSupervisor.start_link/1","type":"function"},{"doc":"🔨🐍\nA brave snake worker slithering across...\n\nThis module defines a GenServer that manages a snake worker, which interacts with a Python process to execute specified functions asynchronously.\nThe main :run_snake call, creates a `Task.async/1` which calls python and handles exceptions returning python result or an Error struct which gets sent with signal to the caller process. This `Task` gets awaited inside the :run cast(). The original call() returns :ok","ref":"Venomous.SnakeWorker.html","title":"Venomous.SnakeWorker","type":"module"},{"doc":"- Starts and initializes a Python process.\n- Executes Python functions with given arguments.\n- Handles the results of the Python function calls, including errors.\n- Returns the result with a signal to caller process","ref":"Venomous.SnakeWorker.html#module-features","title":"Features - Venomous.SnakeWorker","type":"module"},{"doc":"- On initialization, the worker starts a Python process.\n- If provided, the worker initializes an encoder by calling a specified Python function with arguments.\n- The worker can run Python functions on demand and return the results to the caller.","ref":"Venomous.SnakeWorker.html#module-process-lifecycle","title":"Process Lifecycle - Venomous.SnakeWorker","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Venomous.SnakeWorker.html#child_spec/1","title":"Venomous.SnakeWorker.child_spec/1","type":"function"},{"doc":"","ref":"Venomous.SnakeWorker.html#handle_continue/2","title":"Venomous.SnakeWorker.handle_continue/2","type":"function"},{"doc":"","ref":"Venomous.SnakeWorker.html#init/1","title":"Venomous.SnakeWorker.init/1","type":"function"},{"doc":"","ref":"Venomous.SnakeWorker.html#start_link/1","title":"Venomous.SnakeWorker.start_link/1","type":"function"},{"doc":"![Venomous](https://github.com/RustySnek/Venomous/blob/master/assets/venomous_logo.png)\n\n> A wrapper for managing concurrent [Erlport](http://erlport.org/) Python processes with ease.\n\n[![CI](https://github.com/rustysnek/venomous/actions/workflows/elixir.yml/badge.svg)](https://github.com/rustysnek/venomous/actions/workflows/elixir.yml)\n[![Hex.pm](https://img.shields.io/hexpm/v/venomous)](https://hex.pm/packages/venomous)\n[![Hex.pm](http://img.shields.io/hexpm/dt/venomous.svg)](https://hex.pm/packages/venomous)\n\nVenomous is a wrapper around erlport python Ports, designed to simplify concurrent use. It focuses on dynamic extensibility, like spawning, reusing and killing processes on demand. Furthermore, unused processes get automatically killed by scheduled process which can be configured inside config.exs. Venomous core functions capture and handle :EXIT calls ensuring that all python process die with it and do not continue their execution.","ref":"readme.html","title":"README","type":"extras"},{"doc":"Add `:venomous` to your list of dependencies in `mix.exs`:\n```elixir\ndef deps do\n  [\n    {:venomous, \"~> 0.3.0\"}\n  ]\nend\n```","ref":"readme.html#installation","title":"Installation - README","type":"extras"},{"doc":"Check the [documentation](https://hexdocs.pm/venomous) for more in-depth information.\n  \n  For custom type conversion see the [Handling Erlport API](https://github.com/RustySnek/Venomous/blob/master/PYTHON.md)","ref":"readme.html#getting-started","title":"Getting Started - README","type":"extras"},{"doc":"```elixir\n  config :venomous, :snake_manager, %{\n    # Optional :erlport encoder/decoder for type conversion between elixir/python applied to all workers. The function may also include any :erlport callbacks from python api\n    erlport_encoder: %{\n      module: :my_encoder_module,\n      func: :encode_my_snakes_please,\n      args: []\n    },\n    # TTL whenever python process is inactive. Default: 15\n    snake_ttl_minutes: 10,\n    # Number of python workers that don't get cleared by SnakeManager when their TTL while inactive ends. Default: 10\n    perpetual_workers: 1,\n    # Interval for killing python processes past their ttl while inactive. Default: 60_000ms (1 min)\n    cleaner_interval: 5_000\n  }\n  ```","ref":"readme.html#configure-the-snakemanager-options","title":"Configure the SnakeManager options - README","type":"extras"},{"doc":"```elixir\n  defmodule YourApp.Application do\n    @moduledoc false\n\n    use Application\n\n    @doc false\n    def start(_type, _args) do\n      children = [\n        {Venomous.SnakeSupervisor, [strategy: :one_for_one, max_restarts: 0, max_children: 50]}\n      ]\n      opts = [strategy: :one_for_one, name: YourApp.Supervisor]\n      Supervisor.start_link(children, opts)\n    end\n  end\n  ```","ref":"readme.html#configure-the-snakesupervisor-to-start-on-application-boot","title":"Configure the SnakeSupervisor to start on application boot. - README","type":"extras"},{"doc":"","ref":"readme.html#quickstart","title":"Quickstart - README","type":"extras"},{"doc":"```elixir\nalias Venomous.SnakeArgs\nimport Venomous\n\ntimeout = 1_000\nargs = SnakeArgs.from_params(:builtins, :sum, [[0,1,2,3,4,5]])\n\ncase python(args, python_timeout: timeout) do\n    {:retrieve_error, msg} -> \"No Snakes? #{inspect(msg)}\"\n    %{error: :timeout} -> \"We timed out...\"\n    sum -> assert sum == 15\nend\n\n# or just use python!/3 which waits for the available snake.\ntimeout = :infinity\nassert python!(args, python_timeout: timeout) == 15\n```","ref":"readme.html#basic-way-to-call-python-process","title":"Basic way to call python process - README","type":"extras"},{"doc":"Venomous is designed with concurrency, as well as proper exits in mind.\n```elixir\nalias Venomous.SnakeArgs\nimport Venomous\n\n# Venomous can handle as much concurrent python as you've setup\n# in your snake_manager configuration. However the python! will\n# wait for any process to free up in case none are available.\nargs = SnakeArgs.from_params(:time, :sleep, [0.5])\nEnum.map(1..100, fn _ -> \n    Task.async(fn ->\n        python!(args)\n    end)\nend) |> Task.await_many(5_000)\n\n# You can view the spawned and ready snakes using the list_alive_snakes() \nlist_alive_snakes() |> dbg\n```\n```elixir\nalias Venomous.SnakeArgs\nimport Venomous\n\n# Venomous kills the OS pid of the python process on :EXIT\n# ensuring the process will not proceed with the execution further\nEnum.map(1..200, fn _ ->\n  {:ok, pid} =\n    Task.start(fn ->\n      SnakeArgs.from_params(:time, :sleep, [1000]) |> python!()\n    end)\n\n  pid\nend)\n|> Enum.each(fn pid ->\n  Process.send_after(pid, {:EXIT, :snake_slithered_away}, 100)\nend)\n\n# We'll sleep to make sure all exits got sent.\nProcess.sleep(1_000)\nassert list_alive_snakes() == []\n```","ref":"readme.html#concurrency-and-exit-signals","title":"Concurrency and :EXIT signals - README","type":"extras"},{"doc":"# Quick guide on erlport Python API\nYou can read more about this in the [erlport documentation](http://erlport.org/docs/python.html)\n\nYou can find the default data types mapping [here](http://erlport.org/docs/python.html#data-types-mapping)","ref":"python.html","title":"Quick guide on erlport Python API","type":"extras"},{"doc":"The function passed to the snake_manager's erlport_encoder should be a main function containing everything you will setup further.\n```python\nfrom erlport.erlang import set_decoder, set_encoder, set_message_handler\n\ndef main_encoder():\n  set_decoder(decoder_func)\n  set_encoder(encoder_func)\n  set_message_handler(cast_handler_func)\n  return Atom(\"ok\".encode(\"utf-8\"))\n```","ref":"python.html#creating-an-encoder-decoder","title":"Creating an encoder/decoder - Quick guide on erlport Python API","type":"extras"},{"doc":"Here is an example of encoding a simple class\n> While this is possible, a better way would be to handle them in a function that returns this type directly.\n```python\nfrom dataclasses import dataclass\nfrom erlport.erlterms import Atom\n\n@dataclass\nclass Cat:\n  name: str\n  color: str\n  favorite_snacks: list[str]\n\n# By default erlport converts regular strings into charlists.\n# We can handle most of the cases by encoding strings into utf-8 with a simple function like this.\ndef encode_basic_type_strings(data):\n    if isinstance(data, str):\n        return data.encode(\"utf-8\")\n    elif isinstance(data, list):\n        return [encode_basic_type_strings(item) for item in data]\n    elif isinstance(data, tuple):\n        return tuple(encode_basic_type_strings(item) for item in data)\n    elif isinstance(data, dict):\n        return {key: encode_basic_type_strings(value) for key, value in data.items()}\n    else:\n        return data\n\ndef encoder(value: any):\n  if isinstance(value, Cat):\n    # We .__dict__ the class and normalize it's key,values.\n    # Erlport will convert this dict into a Map\n    return {\n      Atom(encode_basic_type_strings(key)):   # Convert the keys to atoms\n      encode_basic_type_strings(val)          # utf-8 Encode strings inside\n      for key, val in value.__dict__.items()\n    }\n  # If none matches we encode strings, and return\n  return encode_basic_type_strings(value)\n```","ref":"python.html#an-encoder-is-a-function-that-handles-conversion-python-elixir","title":"An Encoder is a function that handles conversion PYTHON -> ELIXIR - Quick guide on erlport Python API","type":"extras"},{"doc":"Here is an example of decoding elixir's parameters\n```python\nfrom erlport.erlterms import Atom, Map\n\ndef decoder(value: any):\n  # Elixir strings convert to bytes, we can decode them into utf-8 strings.\n  if isinstance(value, bytes):\n    return value.decode('utf-8')\n  if isistance(value, Map):\n      # If its a Map custom type we decode bytes into utf-8 strings\n      return {\n            key.decode(\"utf-8\"): [v.decode(\"utf-8\") for v in val]\n            for key, val in value.items()\n      }\n  # if none get caught we just return the raw inputs\n  return value\n\n```","ref":"python.html#a-decoder-is-a-function-that-handles-conversion-elixir-python","title":"A decoder is a function that handles conversion ELIXIR -> PYTHON - Quick guide on erlport Python API","type":"extras"}],"producer":{"name":"ex_doc","version":[48,46,51,52,46,48]}}